package org.artifactory.schedule.monitor;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.Lists;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.concurrent.BasicThreadFactory;
import org.apache.http.StatusLine;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.artifactory.common.ConstantValues;
import org.artifactory.config.InternalCentralConfigService;
import org.artifactory.repo.RemoteRepo;
import org.artifactory.spring.Reloadable;
import org.artifactory.util.BasicHttpUtils;
import org.jetbrains.annotations.NotNull;
import org.jfrog.client.util.PathUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * @author Alexei Vainshtein
 */
@Slf4j
@Service
@Reloadable(beanClass = OnlineMonitorService.class,
        initAfter = InternalCentralConfigService.class)
public class OnlineMonitorServiceImpl implements OnlineMonitorService {

    @VisibleForTesting
    final Map<String, ScheduledFuture> offlineReposFutures = new ConcurrentHashMap<>();
    @VisibleForTesting
    ScheduledExecutorService onlineMonitorScheduler;
    private static final List<Integer> offlineStatusCodes = Lists.newArrayList(502, 503, 504, 505);

    @Autowired
    public OnlineMonitorServiceImpl() {
        initOnlineMonitorScheduler();
    }

    @Override
    public void createOfflineRepoMonitor(RemoteRepo repo) {
        if (offlineReposFutures.containsKey(repo.getKey())) {
            log.debug("Found existing offline task for {}. No need to add another", repo.getKey());
            return;
        }
        log.info("Creating online scheduler for {}", repo.getKey());
        OnlineMonitorRunnable runnable = createRunnable(repo);
        ScheduledFuture<?> scheduledFuture =
                onlineMonitorScheduler
                        .schedule(runnable, 0, TimeUnit.MILLISECONDS);
        ScheduledFuture oldScheduled = offlineReposFutures.put(repo.getKey(), scheduledFuture);
        // Checks if simultaneously were added two tasks. If so, we can cancel the other
        if (oldScheduled != null) {
            log.debug("Removing previous scheduled task for {}", repo.getKey());
            oldScheduled.cancel(true);
        }
    }

    @NotNull
    OnlineMonitorRunnable createRunnable(RemoteRepo repo) {
        return new OnlineMonitorRunnable(repo);
    }

    void initOnlineMonitorScheduler() {
        if (this.onlineMonitorScheduler == null) {
            var threadsCount = ConstantValues.onlineMonitorThreadsCount.getInt();
            ScheduledExecutorService executorService = Executors.newScheduledThreadPool(threadsCount,
                    new BasicThreadFactory.Builder()
                            .namingPattern("online-monitor-%d")
                            .daemon(true)
                            .build());
            ((ScheduledThreadPoolExecutor) executorService).setKeepAliveTime(60, TimeUnit.SECONDS);
            ((ScheduledThreadPoolExecutor) executorService).allowCoreThreadTimeOut(true);
            ((ScheduledThreadPoolExecutor) executorService).setRemoveOnCancelPolicy(true);
            ((ScheduledThreadPoolExecutor) executorService).setMaximumPoolSize(
                    threadsCount);
            this.onlineMonitorScheduler = executorService;
        }
    }

    @Override
    public final boolean isResourceUnavailable(StatusLine status) {
        return offlineStatusCodes.contains(status.getStatusCode());
    }

    @Override
    public void stopMonitorForRepo(RemoteRepo repo) {
        terminateOldTaskIfNeeded(repo.getKey());
    }

    @Override
    public void destroy() {
        if (this.onlineMonitorScheduler != null) {
            this.onlineMonitorScheduler.shutdown();
        }
    }

    /**
     * Removes and cancel the running task
     * @param key - The repo key to stop and remove the monitor
     */
    void terminateOldTaskIfNeeded(String key) {
        ScheduledFuture onlineMonitorFuture = removeRepoFromOfflineMap(key);
        if (onlineMonitorFuture != null) {
            log.debug("Canceling previous task for {}", key);
            onlineMonitorFuture.cancel(true);
        }
    }

    /**
     * Removes the repo from the offline map
     * @param key - repo name
     * @return the scheduled task of the repo
     */
    ScheduledFuture removeRepoFromOfflineMap(String key) {
        return offlineReposFutures.remove(key);
    }

    @VisibleForTesting
    class OnlineMonitorRunnable implements Runnable {
        /**
         * max attempts until reaching the maximum wait time
         */
        private static final int MAX_FAILED_ATTEMPTS = 10;

        /**
         * Failed requests counter
         */
        private int failedAttempts = 0;
        private final RemoteRepo repo;

        public OnlineMonitorRunnable(RemoteRepo repo) {
            this.repo = repo;
        }

        @Override
        public void run() {
            String key = repo.getKey();
            log.debug("Online monitor started for {}", key);
            if (checkOnline()) {
                if (repo.isAssumedOffline()) {
                    log.info("{} is back online!", key);
                }
                repo.resetAssumedOffline();
                // Task is back online, thread will end and need to remove the map
                removeRepoFromOfflineMap(key);
                return;
            }
            if (!repo.isAssumedOffline()) {
                log.info("{} is inaccessible. Setting as offline!", key);
                repo.setAssumedOffline();
            }
            long nextOnlineCheckDelay = calculateNextOnlineCheckDelay();
            ScheduledFuture<?> scheduledFuture =
                    onlineMonitorScheduler
                            .schedule(this, nextOnlineCheckDelay, TimeUnit.MILLISECONDS);
            offlineReposFutures.put(key, scheduledFuture);
            log.debug("Online monitor for repo {} will wait for {} millis", key, nextOnlineCheckDelay);
        }

        /**
         * Calculates the next time the check for online need to happen.
         * This will increase exponentially.
         * @return the time until the next request
         */
        long calculateNextOnlineCheckDelay() {
            long maxFailureCacheSecs = repo.getAssumedOfflinePeriodSecs();
            long maxFailureCacheMillis = TimeUnit.SECONDS.toMillis(maxFailureCacheSecs);    // always >= 1000

            long nextOnlineCheckDelayMillis;
            failedAttempts++;
            if (failedAttempts < MAX_FAILED_ATTEMPTS) {
                if (maxFailureCacheSecs / MAX_FAILED_ATTEMPTS < 2) {
                    long failurePenaltyMillis = maxFailureCacheMillis / MAX_FAILED_ATTEMPTS;
                    nextOnlineCheckDelayMillis = failedAttempts * failurePenaltyMillis;
                } else {
                    // exponential delay
                    // calculate the base of the exponential equation based on the MAX_FAILED_ATTEMPTS and max
                    //offline period
                    // BASE pow MAX_FAILED_ATTEMPTS = MAX_DELAY ==> BASE = MAX_DELAY pow 1/MAX_FAILED_ATTEMPTS
                    double base = Math.pow(maxFailureCacheMillis, 1.0 / (double) MAX_FAILED_ATTEMPTS);
                    nextOnlineCheckDelayMillis = (long) Math.pow(base, failedAttempts);
                    // in any case don't attempt too rapidly
                    nextOnlineCheckDelayMillis = Math.max(100, nextOnlineCheckDelayMillis);
                }
            } else {
                nextOnlineCheckDelayMillis = maxFailureCacheMillis;
            }
            return nextOnlineCheckDelayMillis;
        }

        /**
         * Checks if the given url back online
         * @return - true if online, false otherwise
         */
        boolean checkOnline() {
            // always test with url trailing slash
            String url = PathUtils.addTrailingSlash(repo.getUrl());
            HttpGet getMethod = new HttpGet(BasicHttpUtils.encodeQuery(url));
            try (CloseableHttpResponse response = repo.getHttpClient().execute(getMethod)) {
                log.debug("Online monitor checking URL: {}", url);
                StatusLine status = response.getStatusLine();
                log.debug("Online monitor http method completed with no exception: {}: {}",
                        status.getStatusCode(), status.getReasonPhrase());

                if (isResourceUnavailable(status)) { //RTFACT-6528
                    log.debug("Considering repo {} as offline due to the return code '{}'", repo.getKey(),
                            status.getStatusCode());
                } else {
                    return true;
                }
            } catch (IllegalStateException | IOException e) {
                // IllegalStateException can be thrown while this tread in race with destroy()
                // i.e. closed client is accesses before interrupt() takes effect
                log.debug("Online monitor http method failed: {}: {}", e.getClass().getName(), e.getMessage());
            }
            return false;
        }
    }
}
