package org.artifactory.ph;

import com.fasterxml.jackson.core.type.TypeReference;
import com.google.common.collect.LinkedHashMultiset;
import org.artifactory.api.events.EventLogHintManager;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.api.rest.restmodel.JsonUtil;
import org.artifactory.common.MutableStatusHolder;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.listener.PackageArtifactEventListener;
import org.artifactory.core.api.ph.listener.PackageEventListenerContext;
import org.artifactory.core.api.ph.model.PackageArtifact;
import org.artifactory.core.api.ph.service.PackageSpecificationAdapter;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.event.EventHintType;
import org.artifactory.fs.FileInfo;
import org.artifactory.fs.MutableFileInfo;
import org.artifactory.fs.RepoResource;
import org.artifactory.md.Properties;
import org.artifactory.repo.HttpRepo;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.resource.RepoResourceInfo;
import org.artifactory.sapi.fs.VfsFile;
import org.artifactory.sapi.interceptor.context.BulkCopyMoveContext;
import org.artifactory.sapi.interceptor.context.DeleteContext;
import org.artifactory.sapi.interceptor.context.InterceptorCreateContext;
import org.artifactory.sapi.interceptor.context.InterceptorMoveCopyContext;
import org.artifactory.storage.db.fs.model.DbMutableFile;
import org.artifactory.util.StringInputStream;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.Spy;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;
import static org.testng.Assert.assertEquals;

@Test
public class PackageStorageInterceptorImplTest {

    private static final RepoPath LOCAL_REPO_PATH = RepoPathFactory.create("local_repo1", "a/b/c/f1");
    private static final RepoPath REMOTE_REPO_PATH = RepoPathFactory.create("remote_repo1", "a/b/c/f1");

    private PackageStorageInterceptorImpl interceptor;

    @Spy
    private PackageArtifactEventListener packagesListener;
    @Spy
    private PackageSpecificationAdapter packageSpecificationAdapter;
    @Spy
    private InternalRepositoryService repositoryService;
    @Mock
    private DbMutableFile file;
    @Mock
    private RepoDescriptor descriptor;
    @Mock
    private MutableStatusHolder statusHolder;
    @Mock
    private Properties properties;
    @Mock
    private InterceptorMoveCopyContext interceptorMoveCopyContext;

    @BeforeMethod
    public void setup() throws PackageException {
        openMocks(this);
        interceptor =
                new PackageStorageInterceptorImpl(packagesListener, repositoryService, packageSpecificationAdapter);
        setup(file, LOCAL_REPO_PATH);
        Repo<RepoDescriptor> localRepo = mock(Repo.class);
        Repo<RepoDescriptor> remoteRepo = mock(Repo.class);
        when(descriptor.getType()).thenReturn(RepoType.CRAN);
        when(localRepo.getDescriptor()).thenReturn(descriptor);
        when(localRepo.isLocal()).thenReturn(true);
        when(localRepo.isCache()).thenReturn(false);
        when(remoteRepo.getDescriptor()).thenReturn(descriptor);
        when(remoteRepo.isLocal()).thenReturn(false);
        when(remoteRepo.isCache()).thenReturn(false);

        when(repositoryService.repositoryByKey(eq(LOCAL_REPO_PATH.getRepoKey()))).thenReturn(localRepo);
        when(repositoryService.repositoryByKey(eq(REMOTE_REPO_PATH.getRepoKey()))).thenReturn(remoteRepo);
        when(packageSpecificationAdapter.isSupportedPackage(anyString())).thenReturn(true);
    }

    void setup(VfsFile<?> file, RepoPath repoPath) {
        when(file.length()).thenReturn(Long.MAX_VALUE);
        when(file.getSha1()).thenReturn("sha1");
        when(file.getSha2()).thenReturn("sha2");
        when(file.getMd5()).thenReturn("md5");
        when(file.isFolder()).thenReturn(false);
        when(file.isFile()).thenReturn(true);
        when(file.getRepoKey()).thenReturn(repoPath.getRepoKey());
        FileInfo fileInfo = Mockito.mock(MutableFileInfo.class);
        when(file.getInfo()).thenReturn(fileInfo);
        when(fileInfo.getLastModified()).thenReturn(System.currentTimeMillis());
        when(fileInfo.getCreated()).thenReturn(System.currentTimeMillis());
        when(fileInfo.getLastUpdated()).thenReturn(System.currentTimeMillis());
        when(fileInfo.getModifiedBy()).thenReturn("System");
        when(fileInfo.getCreatedBy()).thenReturn("System");


        when(properties.keys()).thenReturn(LinkedHashMultiset.create(Set.of()));
        doAnswer(invocation -> properties).when(file).getProperties();

        when(file.getInfo()).thenReturn(fileInfo);

        when(fileInfo.getSha1()).thenReturn("sha1");
        when(fileInfo.getSha2()).thenReturn("sha2");
        when(fileInfo.getMd5()).thenReturn("md5");
        when(fileInfo.getRepoPath()).thenReturn(repoPath);
        when(fileInfo.getCreated()).thenReturn(Long.MAX_VALUE);
        when(fileInfo.getRepoKey()).thenReturn(repoPath.getRepoKey());
        when(fileInfo.getLastModified()).thenReturn(Long.MAX_VALUE);
        when(fileInfo.getName()).thenReturn(repoPath.getName());
        when(fileInfo.getRelPath()).thenReturn(repoPath.getPath());
        when(fileInfo.getModifiedBy()).thenReturn("admin");
        when(fileInfo.getLastUpdated()).thenReturn(Long.MAX_VALUE);
        when(fileInfo.getCreatedBy()).thenReturn("admin");
    }

    @Test
    public void testNoRepository() {
        when(repositoryService.repositoryByKey(eq(LOCAL_REPO_PATH.getRepoKey()))).thenReturn(null);

        interceptor.beforeCreate(file, statusHolder);

        verify(packagesListener, never()).beforeCreate(any(PackageArtifact.class));
    }

    @Test
    public void testUnsupportedPackageType() {

        when(descriptor.getType()).thenReturn(RepoType.Support);

        interceptor.beforeCreate(file, statusHolder);

        verify(packagesListener, never()).beforeCreate(any(PackageArtifact.class));
    }

    @Test
    public void testAfterCreate() {

        interceptor.afterCreate(file, statusHolder, mock(InterceptorCreateContext.class));

        ArgumentCaptor<PackageArtifact> artifact = ArgumentCaptor.forClass(PackageArtifact.class);

        verify(packagesListener, times(1)).afterCreate(artifact.capture());

        assertEquals(artifact.getValue().getRepoKey(), LOCAL_REPO_PATH.getRepoKey());

    }

    @Test
    public void testBeforeDelete() throws IOException {

        EventLogHintManager.put(EventHintType.PackageIndexExtraInfo, JsonUtil.jsonToString(Map.of(
                "k0", "v0",
                "k1", "v1")));

        doAnswer(invocation -> {
            PackageEventListenerContext context = invocation.getArgument(1);
            context.setEventExtraInfo(Map.of("k1", "v11", "k2", 2));
            return null;
        }).when(packagesListener)
                .beforeDelete(any(PackageArtifact.class), any(PackageEventListenerContext.class));

        interceptor.beforeDelete(file, statusHolder, mock(DeleteContext.class));
        ArgumentCaptor<PackageArtifact> artifact = ArgumentCaptor.forClass(PackageArtifact.class);
        ArgumentCaptor<PackageEventListenerContext> context = ArgumentCaptor.forClass(
                PackageEventListenerContext.class);
        verify(packagesListener, times(1)).beforeDelete(artifact.capture(), context.capture());

        assertEquals(artifact.getValue().getRepoKey(), LOCAL_REPO_PATH.getRepoKey());

        String hintValue = EventHintType.getValue(EventLogHintManager.getHints(), EventHintType.PackageIndexExtraInfo);
        Map<String, Object> extraInfo = JsonUtil.mapDataToModel(new StringInputStream(hintValue),
                new TypeReference<>() {
                });

        assertEquals(extraInfo.size(), 3);
        assertEquals(extraInfo.get("k0"), "v0");
        assertEquals(extraInfo.get("k1"), "v11");
        assertEquals(extraInfo.get("k2"), 2);
    }

    @Test
    public void testAfterDelete() {

        interceptor.afterDelete(file, statusHolder, mock(DeleteContext.class));

        ArgumentCaptor<PackageArtifact> artifact = ArgumentCaptor.forClass(PackageArtifact.class);

        verify(packagesListener, times(1)).afterDelete(artifact.capture());

        assertEquals(artifact.getValue().getRepoKey(), LOCAL_REPO_PATH.getRepoKey());

    }

    @Test
    public void testAfterMove() {

        RepoPath targetPath = RepoPathFactory.create("repo2", LOCAL_REPO_PATH.getPath());
        VfsFile<?> targetFile = mock(VfsFile.class);
        setup(targetFile, targetPath);

        Repo<RepoDescriptor> targetRepo = mock(Repo.class);
        RepoDescriptor targetRepoDescriptor = mock(RepoDescriptor.class);
        when(targetRepoDescriptor.getType()).thenReturn(RepoType.CRAN);
        when(targetRepo.getDescriptor()).thenReturn(targetRepoDescriptor);
        when(repositoryService.repositoryByKey(eq(targetPath.getRepoKey()))).thenReturn(targetRepo);

        interceptor.afterMove(file, targetFile, statusHolder, properties, interceptorMoveCopyContext);

        ArgumentCaptor<PackageArtifact> source = ArgumentCaptor.forClass(PackageArtifact.class);
        ArgumentCaptor<PackageArtifact> target = ArgumentCaptor.forClass(PackageArtifact.class);

        verify(packagesListener, times(1)).afterMove(source.capture(), target.capture());
        assertEquals(source.getValue().getRepoKey(), LOCAL_REPO_PATH.getRepoKey());
        assertEquals(target.getValue().getRepoKey(), targetPath.getRepoKey());


    }

    @Test
    public void testAfterCopy() {

        RepoPath targetPath = RepoPathFactory.create("repo2", LOCAL_REPO_PATH.getPath());
        VfsFile<?> targetFile = mock(VfsFile.class);
        setup(targetFile, targetPath);

        Repo<RepoDescriptor> targetRepo = mock(Repo.class);
        RepoDescriptor targetRepoDescriptor = mock(RepoDescriptor.class);
        when(targetRepoDescriptor.getType()).thenReturn(RepoType.CRAN);
        when(targetRepo.getDescriptor()).thenReturn(targetRepoDescriptor);
        when(repositoryService.repositoryByKey(eq(targetPath.getRepoKey()))).thenReturn(targetRepo);

        interceptor.afterCopy(file, targetFile, statusHolder, properties, interceptorMoveCopyContext);

        ArgumentCaptor<PackageArtifact> source = ArgumentCaptor.forClass(PackageArtifact.class);
        ArgumentCaptor<PackageArtifact> target = ArgumentCaptor.forClass(PackageArtifact.class);

        verify(packagesListener, times(1)).afterCopy(source.capture(), target.capture());
        assertEquals(source.getValue().getRepoKey(), LOCAL_REPO_PATH.getRepoKey());
        assertEquals(target.getValue().getRepoKey(), targetPath.getRepoKey());

    }

    @Test
    public void testAfterPropertyCreate() {
        interceptor.afterPropertyCreate(file, statusHolder, "k1", "v1", "v2");

        ArgumentCaptor<PackageArtifact> artifact = ArgumentCaptor.forClass(PackageArtifact.class);
        ArgumentCaptor<String> key = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<Boolean> delete = ArgumentCaptor.forClass(Boolean.class);
        ArgumentCaptor<String[]> values = ArgumentCaptor.forClass(String[].class);

        verify(packagesListener, times(1)).afterPropertyChange(artifact.capture(), key.capture(), delete.capture(),
                values.capture());

        assertEquals(artifact.getValue().getRepoKey(), LOCAL_REPO_PATH.getRepoKey());
        assertEquals(key.getValue(), "k1");
        assertEquals(delete.getValue(), Boolean.FALSE);
        assertEquals(values.getAllValues(), Arrays.asList("v1", "v2"));
    }

    @Test
    public void testAfterPropertyDelete() {
        interceptor.afterPropertyDelete(file, statusHolder, "k1");

        ArgumentCaptor<PackageArtifact> artifact = ArgumentCaptor.forClass(PackageArtifact.class);
        ArgumentCaptor<String> key = ArgumentCaptor.forClass(String.class);
        ArgumentCaptor<Boolean> delete = ArgumentCaptor.forClass(Boolean.class);
        ArgumentCaptor<String[]> values = ArgumentCaptor.forClass(String[].class);

        verify(packagesListener, times(1)).afterPropertyChange(artifact.capture(), key.capture(), delete.capture(),
                values.capture());

        assertEquals(artifact.getValue().getRepoKey(), LOCAL_REPO_PATH.getRepoKey());
        assertEquals(key.getValue(), "k1");
        assertEquals(delete.getValue(), Boolean.TRUE);
        assertEquals(values.getAllValues(), List.of());
    }

    @Test
    public void testAfterRemoteDownload() throws RepoRejectException {
        HttpRepo repo = mock(HttpRepo.class);
        HttpRepoDescriptor repoDescriptor = mock(HttpRepoDescriptor.class);
        when(repo.isLocal()).thenReturn(false);
        when(repo.isCache()).thenReturn(false);
        when(repo.getDescriptor()).thenReturn(repoDescriptor);
        when(repoDescriptor.getType()).thenReturn(RepoType.CRAN);
        when(repositoryService.remoteRepositoryByKey(eq(REMOTE_REPO_PATH.getRepoKey()))).thenReturn(repo);
        RepoResource repoResource = mock(RepoResource.class);
        RepoResourceInfo info = mock(RepoResourceInfo.class);
        when(info.getRepoPath()).thenReturn(REMOTE_REPO_PATH);
        when(info.getLastModified()).thenReturn(Long.MAX_VALUE);
        when(info.getSize()).thenReturn(Long.MAX_VALUE);
        when(info.getSha1()).thenReturn("sha1");
        when(info.getSha2()).thenReturn("sha2");
        when(info.getMd5()).thenReturn("md5");
        when(info.getName()).thenReturn(REMOTE_REPO_PATH.getName());
        when(repoResource.getInfo()).thenReturn(info);
        when(repoResource.getRepoKey()).thenReturn(REMOTE_REPO_PATH.getRepoKey());

        interceptor.afterRemoteDownload(repoResource);

        ArgumentCaptor<PackageArtifact> artifact = ArgumentCaptor.forClass(PackageArtifact.class);

        verify(packagesListener, times(1)).afterRemoteDownload(artifact.capture());

        assertEquals(artifact.getValue().getRepoKey(), REMOTE_REPO_PATH.getRepoKey());
    }

    @Test
    public void testBeforeCopyFiles() {
        interceptor.beforeMoveFiles(mock(BulkCopyMoveContext.class));
        verify(repositoryService, never()).repositoryByKey(anyString());
    }

    @Test
    public void testBeforeMoveFiles() {
        interceptor.beforeMoveFiles(mock(BulkCopyMoveContext.class));
        verify(repositoryService, never()).repositoryByKey(anyString());
    }

    @Test
    public void testAssertDeletePathAllowed() {
        interceptor.assertDeletePathAllowed(LOCAL_REPO_PATH, statusHolder);
        verify(repositoryService, never()).repositoryByKey(anyString());
    }

    @Test
    public void testAssertDeleteRepoAllowed() {
        interceptor.assertDeleteRepoAllowed(LOCAL_REPO_PATH.getRepoKey(), statusHolder);
        verify(repositoryService, never()).repositoryByKey(anyString());
    }
}