package org.artifactory.core.api.ph.service;

import org.artifactory.common.repo.PackageType;
import org.artifactory.common.repo.RepoInfo;
import org.artifactory.core.api.ph.exception.PackageException;
import org.jetbrains.annotations.NotNull;

/**
 * The PackageInfoAdapter handles package type verifications over specific artifacts paths.
 * The Implementation is provided by the package handler (in OSS, a default empty bean is loaded by the core)
 *
 * @author BarakH
 */
public interface PackageSpecificationAdapter {

    /**
     * Indicates whether the specified path is a package in the given repository key.
     *
     * @return true if the artifact path represent a package
     */
    boolean isPackageArtifact(@NotNull String repoKey, @NotNull String path) throws PackageException;

    /**
     * Indicates whether the specified path is metadata path
     *
     * @return true if the path is metadata, false otherwise
     */
    boolean isExpirable(@NotNull PackageType packageType, @NotNull String path) throws PackageException;

    /**
     * Indicates whether the file is generated by artifactory.
     *
     * @return true if the file should not be replicated (either generated or tmp).
     */
    boolean isLocalGenerated(@NotNull PackageType packageType, @NotNull String path) throws PackageException;

    /**
     * Indicates whether the property is locally generated.
     *
     * @return True if the property key can be generated natively when deploying a new package to the repo.
     */
    boolean isLocallyGeneratedProperty(@NotNull PackageType packageType, @NotNull String propertyKey)
            throws PackageException;

    /**
     * The method is used by artifactory to identify whether the package is supported or not by the new framework
     *
     * @param packageType the package type to check
     * @return true if the package is supported
     */
    boolean isSupportedPackage(PackageType packageType) throws PackageException;


    /**
     * The method is used by artifactory to identify whether the package type of the repo is supported or not by the
     * new framework
     *
     * @param repoKey the repoKey type to check
     * @return true if the package type of the repo is supported
     */
    boolean isSupportedPackage(String repoKey) throws PackageException;

    /**
     * The method is used by artifactory to identify whether the package type & repo type are supporting reindex
     */
    boolean hasReindex(RepoInfo repoInfo);
}
