/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.dao.locks;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.Sets;
import lombok.NonNull;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.lang.StringUtils;
import org.artifactory.storage.db.util.querybuilder.ArtifactoryQueryWriter;
import org.artifactory.util.CollectionUtils;
import org.jfrog.storage.DbProperties;
import org.jfrog.storage.DbType;
import org.jfrog.storage.JFrogDataSource;
import org.jfrog.storage.util.DbStatementUtils;
import org.jfrog.storage.util.DbUtils;
import org.jfrog.storage.util.querybuilder.QueryWriter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.stereotype.Repository;

import javax.annotation.Nullable;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.List;
import java.util.Set;

import static org.artifactory.common.ConstantValues.dbLockConditionalInsertEnabled;

/**
 * @author gidis
 * @author Mor Merhav
 */
@Repository
public class DbDistributeLocksDao {

    public static final String TABLE_DISTRIBUTED_LOCKS = "distributed_locks";

    private static final Logger log = LoggerFactory.getLogger(DbDistributeLocksDao.class);

    private static final String ORACLE_CONDITIONAL_INSERT_STATEMENT =
            "MERGE INTO DISTRIBUTED_LOCKS " +
                    "USING (SELECT 1 FROM dual) " +
                    "ON (DISTRIBUTED_LOCKS.category = ? and DISTRIBUTED_LOCKS.lock_key = ?) " +
                    "    WHEN NOT MATCHED THEN " +
                    "        INSERT (category, lock_key, owner, owner_thread, owner_thread_name, acquire_time) " +
                    "        VALUES (?, ?, ?, ?, ?, ?)";

    private static final String POSTGRES_CONDITIONAL_INSERT_STATEMENT =
            "INSERT INTO distributed_locks VALUES(?,?,?,?,?,?) ON CONFLICT DO NOTHING";

    private static final String NON_CONDITIONAL_INSERT_STATEMENT =
            "INSERT INTO distributed_locks VALUES(?,?,?,?,?,?)";

    private final JFrogDataSource uniqueLocksDataSource;
    private final NamedParameterJdbcTemplate jdbcTemplate;
    private final DbType dbType;
    private boolean conditionalInsertEnabled;

    @Autowired
    public DbDistributeLocksDao(@Qualifier("uniqueLockDataSource") JFrogDataSource uniqueLocksDataSource,
            DbProperties dbProperties) {
        this.uniqueLocksDataSource = uniqueLocksDataSource;
        this.jdbcTemplate = new NamedParameterJdbcTemplate(uniqueLocksDataSource);
        this.dbType = dbProperties.getDbType();
        conditionalInsertEnabled = isConditionalInsertEnabled();
    }

    public boolean isLocked(String category, String key) throws SQLException {
        key = normalizeKey(key);
        Connection con = null;
        PreparedStatement stmt = null;
        ResultSet resultSet = null;
        try {
            con = uniqueLocksDataSource.getConnection();
            stmt = con.prepareStatement("select owner from distributed_locks where lock_key=? and category=?");
            stmt.setString(1, key);
            stmt.setString(2, category);
            resultSet = stmt.executeQuery();
            return resultSet.next();
        } finally {
            close(con, stmt, resultSet);
        }
    }

    public boolean isLockedByPrefix(String category, String lockPrefix, long limit) {
        return !getLocksByPrefix(category, lockPrefix, limit).isEmpty();
    }

    @NonNull
    public List<LockInfo> getLocksByPrefix(String category, String lockPrefix, long limit) {
        MapSqlParameterSource sqlParameterSource = new MapSqlParameterSource();
        sqlParameterSource.addValue("category", category);
        sqlParameterSource.addValue("lockprefix", lockPrefix + '%');
        QueryWriter queryWriter = new ArtifactoryQueryWriter();
        String sql = queryWriter
                .select("*")
                .from(TABLE_DISTRIBUTED_LOCKS)
                .where("category = (:category) AND lock_key LIKE (:lockprefix)")
                .limit(limit)
                .build();
        return jdbcTemplate.query(sql, sqlParameterSource, (resultSet, i) -> toLockInfo(resultSet));
    }

    public int lockingMapSize(String category) throws SQLException {
        Connection con = null;
        PreparedStatement stmt = null;
        ResultSet resultSet = null;
        try {
            con = uniqueLocksDataSource.getConnection();
            stmt = con.prepareStatement("select count(*) from distributed_locks where category=?");
            stmt.setString(1, category);
            resultSet = stmt.executeQuery();
            if (resultSet.next()) {
                return resultSet.getInt(1);
            } else {
                return 0;
            }
        } finally {
            close(con, stmt, resultSet);
        }
    }

    public Set<String> lockingMapKeySet(String category) throws SQLException {
        Connection con = null;
        PreparedStatement stmt = null;
        ResultSet resultSet = null;
        Set<String> set = Sets.newHashSet();
        try {
            con = uniqueLocksDataSource.getConnection();
            stmt = con.prepareStatement("select lock_key from distributed_locks where category=?");
            stmt.setString(1, category);
            resultSet = stmt.executeQuery();
            while (resultSet.next()) {
                set.add(resultSet.getString(1));
            }
        } finally {
            close(con, stmt, resultSet);
        }
        return set;
    }

    public boolean unlock(String category, String key, String owner) throws SQLException {
        key = normalizeKey(key);
        Connection con = null;
        PreparedStatement stmt = null;
        try {
            con = uniqueLocksDataSource.getConnection();
            stmt = con.prepareStatement("delete from distributed_locks where lock_key=? and category=? and owner=? " +
                    " and owner_thread=?");
            stmt.setString(1, key);
            stmt.setString(2, category);
            stmt.setString(3, owner);
            stmt.setLong(4, Thread.currentThread().getId());
            int effectedRows = stmt.executeUpdate();
            return effectedRows > 0;
        } finally {
            close(con, stmt);
        }
    }

    /**
     * Tries to acquire lock by conditional row insertion to DistributeLocks table.
     *
     * NOTE: this DOES NOT apply on Derby DB, as this DB isn't supported in HA.
     * meaning you will still get errors in Derby's DB log if insertion of raw failed.
     *
     * @param lockInfo with the lock info to insert to the DB table
     * @return true if succeeded to insert row, otherwise false
     */
    public boolean tryToAcquireLock(LockInfo lockInfo) throws SQLException {
        String key = normalizeKey(lockInfo.getKey());
        Connection con = null;
        PreparedStatement stmt = null;
        try {
            con = uniqueLocksDataSource.getConnection();
            stmt = getPreparedStatementForInsertIfNotExists(con, lockInfo, key);
            return stmt.executeUpdate() == 1;
        } catch (SQLException e) {
            String error = "Failed to insert lock row to the 'distributed_locks' table: {}. {}";
            if (conditionalInsertEnabled && !dbType.equals(DbType.ORACLE)) {
                log.error(error, lockInfo, e.getMessage());
            } else if (log.isDebugEnabled()) {
                log.debug(error, lockInfo, e.getMessage());
            }
            log.debug("Failed to insert lock.", e);
            return false;
        } finally {
            close(con, stmt);
        }
    }

    public boolean deleteLockByCategoryAndKey(String category, String key) throws SQLException {
        key = normalizeKey(key);
        Connection con = null;
        PreparedStatement stmt = null;
        try {
            con = uniqueLocksDataSource.getConnection();
            stmt = con.prepareStatement("delete from distributed_locks where lock_key=? and category=?");
            stmt.setString(1, key);
            stmt.setString(2, category);
            return stmt.executeUpdate() == 1;
        } finally {
            close(con, stmt);
        }
    }

    public Connection getLocksConnection() {
        try {
            return uniqueLocksDataSource.getConnection();
        } catch (SQLException e) {
            throw new IllegalStateException("Unable to get connection from unique lock data source", e);
        }
    }

    public Set<LockInfo> getAllLocksByCategoryNotOwnedBy(Set<String> serverIds, String category) {
        if (CollectionUtils.isNullOrEmpty(serverIds)) {
            return Sets.newHashSet();
        }

        MapSqlParameterSource sqlParameterSource = new MapSqlParameterSource();
        sqlParameterSource.addValue("owners", serverIds);
        sqlParameterSource.addValue("category", category);
        String sql = "SELECT * FROM distributed_locks WHERE owner NOT IN (:owners) AND category=(:category)";

        List<LockInfo> lockInfoList = jdbcTemplate
                .query(sql, sqlParameterSource, (resultSet, i) -> toLockInfo(resultSet));
        return Sets.newHashSet(lockInfoList);
    }

    public Set<LockInfo> getAllLocksNotOwnedByAndNotContainedInCategories(Set<String> excludedOwners,
            Set<String> excludedCategories) {
        if (CollectionUtils.isNullOrEmpty(excludedOwners) || CollectionUtils.isNullOrEmpty(excludedCategories)) {
            return Sets.newHashSet();
        }
        MapSqlParameterSource sqlParameterSource = new MapSqlParameterSource();
        sqlParameterSource.addValue("owners", excludedOwners);
        sqlParameterSource.addValue("categories", excludedCategories);
        String sql = "SELECT * FROM distributed_locks WHERE owner NOT IN (:owners) AND category NOT IN (:categories)";

        List<LockInfo> lockInfoList = jdbcTemplate
                .query(sql, sqlParameterSource, (resultSet, i) -> toLockInfo(resultSet));
        return Sets.newHashSet(lockInfoList);
    }

    public Set<LockInfo> getAllLocksByOwner(String owner) throws SQLException {
        Connection con = null;
        PreparedStatement stmt = null;
        ResultSet resultSet = null;
        Set<LockInfo> set = Sets.newHashSet();
        try {
            con = uniqueLocksDataSource.getConnection();
            stmt = con.prepareStatement(
                    "select category, lock_key, owner_thread, owner_thread_name, acquire_time from distributed_locks " +
                            "where owner=?");
            stmt.setString(1, owner);
            resultSet = stmt.executeQuery();
            while (resultSet.next()) {
                set.add(new LockInfo(resultSet.getString(1), resultSet.getString(2), owner,
                        resultSet.getLong(3), resultSet.getString(4), resultSet.getLong(5)));
            }
        } finally {
            close(con, stmt, resultSet);
        }
        return set;
    }

    public Set<LockInfo> getExpiredLocksNotContainedInCategories(Set<String> excludedCategories,
            long minAllowedTimeMillis) {
        if (CollectionUtils.isNullOrEmpty(excludedCategories)) {
            return Sets.newHashSet();
        }

        MapSqlParameterSource sqlParameterSource = new MapSqlParameterSource();
        sqlParameterSource.addValue("categories", excludedCategories);
        sqlParameterSource.addValue("minAllowedTime", minAllowedTimeMillis);
        String sql = "SELECT * FROM distributed_locks " +
                "WHERE acquire_time < (:minAllowedTime) AND category NOT IN (:categories)";
        List<LockInfo> lockInfoList = jdbcTemplate
                .query(sql, sqlParameterSource, ((resultSet, i) -> toLockInfo(resultSet)));
        return Sets.newHashSet(lockInfoList);
    }

    public Set<LockInfo> getExpiredLocksByCategory(String category, long minAllowedTime) {
        if (StringUtils.isEmpty(category)) {
            return Sets.newHashSet();
        }

        MapSqlParameterSource sqlParameterSource = new MapSqlParameterSource();
        sqlParameterSource.addValue("category", category);
        sqlParameterSource.addValue("minAllowedTime", minAllowedTime);
        String sql = "SELECT * FROM distributed_locks " +
                "WHERE acquire_time < (:minAllowedTime) AND category = (:category)";
        List<LockInfo> lockInfoList = jdbcTemplate
                .query(sql, sqlParameterSource, ((resultSet, i) -> toLockInfo(resultSet)));
        return Sets.newHashSet(lockInfoList);
    }

    public Set<LockInfo> getAllLocksByCategoryAndOwner(String category, String owner) throws SQLException {
        Set<LockInfo> lockInfoSet = Sets.newHashSet();
        Connection con = null;
        PreparedStatement stmt = null;
        ResultSet resultSet = null;
        try {
            con = uniqueLocksDataSource.getConnection();
            String sql =
                    "select category, owner, lock_key, owner_thread, owner_thread_name, acquire_time from " +
                            "distributed_locks where owner=? and category=?";
            stmt = con.prepareStatement(sql);
            stmt.setString(1, owner);
            stmt.setString(2, category);
            resultSet = stmt.executeQuery();
            while (resultSet.next()) {
                lockInfoSet.add(toLockInfo(resultSet));
            }
        } finally {
            close(con, stmt, resultSet);
        }
        return lockInfoSet;
    }

    public LockInfo getLockInfo(String category, String key) throws SQLException {
        key = normalizeKey(key);
        Connection con = null;
        PreparedStatement stmt = null;
        ResultSet resultSet = null;
        try {
            con = uniqueLocksDataSource.getConnection();
            stmt = con.prepareStatement(
                    "select category, lock_key, owner, owner_thread, owner_thread_name, acquire_time from " +
                            "distributed_locks where lock_key=? and category=?");
            stmt.setString(1, key);
            stmt.setString(2, category);
            resultSet = stmt.executeQuery();
            if (resultSet.next()) {
                return toLockInfo(resultSet);
            }
            return null;
        } finally {
            close(con, stmt, resultSet);
        }
    }

    private void close(@Nullable Connection con, @Nullable Statement stmt) {
        close(con, stmt, null);
    }

    private void close(@Nullable Connection con, @Nullable Statement stmt, @Nullable ResultSet rs) {
        try {
            DbUtils.close(rs);
        } finally {
            try {
                DbStatementUtils.close(stmt);
            } finally {
                if (con != null) {
                    try {
                        con.close();
                    } catch (SQLException e) {
                        log.trace("Could not close JDBC connection", e);
                    } catch (Exception e) {
                        log.trace("Unexpected exception when closing JDBC connection", e);
                    }
                }
            }
        }
    }

    private String normalizeKey(String key) {
        try {
            if (StringUtils.isBlank(key)) {
                key = ".";
            }
            if (key.length() > 255) {
                final MessageDigest messageDigest = MessageDigest.getInstance("SHA-1");
                messageDigest.reset();
                messageDigest.update(key.getBytes(StandardCharsets.UTF_8));
                final byte[] resultByte = messageDigest.digest();
                return new String(Hex.encodeHex(resultByte));
            } else {
                return key;
            }
        } catch (Exception e) {
            throw new RuntimeException("Failed to transform lock key into SHA-1", e);
        }
    }

    private LockInfo toLockInfo(ResultSet resultSet) throws SQLException {
        return new LockInfo(
                resultSet.getString("category"),
                resultSet.getString("lock_key"),
                resultSet.getString("owner"),
                resultSet.getLong("owner_thread"),
                resultSet.getString("owner_thread_name"),
                resultSet.getLong("acquire_time")
        );
    }

    @NonNull
    private PreparedStatement getPreparedStatementForInsertIfNotExists(@NonNull Connection con,
            @NonNull LockInfo lockInfo, @NonNull String key) throws SQLException {
        PreparedStatement stmt;
        if (!conditionalInsertEnabled){
            stmt = con.prepareStatement(NON_CONDITIONAL_INSERT_STATEMENT);
            populatePrepStmtGeneralDB(stmt, lockInfo, key);
            return stmt;
        }
        switch (dbType) {
            case POSTGRESQL:
                stmt = con.prepareStatement(POSTGRES_CONDITIONAL_INSERT_STATEMENT);
                populatePrepStmtGeneralDB(stmt, lockInfo, key);
                break;
            case ORACLE:
                stmt = con.prepareStatement(ORACLE_CONDITIONAL_INSERT_STATEMENT);
                populatePrepStmtOracle(stmt, lockInfo, key);
                break;
            case MYSQL:
            case MARIADB:
            case MSSQL:
            case DERBY:  // no conditional
            default:
                stmt = con.prepareStatement(NON_CONDITIONAL_INSERT_STATEMENT);
                populatePrepStmtGeneralDB(stmt, lockInfo, key);
        }
        return stmt;
    }

    @NonNull
    private void populatePrepStmtGeneralDB(@NonNull PreparedStatement stmt,
            @NonNull LockInfo lockInfo, @NonNull String key) throws SQLException {
        stmt.setString(1, lockInfo.getCategory());
        stmt.setString(2, key);
        stmt.setString(3, lockInfo.getOwner());
        stmt.setLong(4, lockInfo.getThreadId());
        stmt.setString(5, lockInfo.getThreadName());
        stmt.setLong(6, lockInfo.getStartedTime());
    }

    @NonNull
    private void populatePrepStmtOracle(@NonNull PreparedStatement stmt,
            @NonNull LockInfo lockInfo, @NonNull String key) throws SQLException {
        stmt.setString(1, lockInfo.getCategory());
        stmt.setString(2, key);
        stmt.setString(3, lockInfo.getCategory());
        stmt.setString(4, key);
        stmt.setString(5, lockInfo.getOwner());
        stmt.setLong(6, lockInfo.getThreadId());
        stmt.setString(7, lockInfo.getThreadName());
        stmt.setLong(8, lockInfo.getStartedTime());
    }

    /**
     * Should not be used directly!
     */
    public boolean isConditionalInsertEnabled() {
        // skip non supported dbs
        if (DbType.DERBY.equals(dbType) || DbType.MSSQL.equals(dbType)
                || DbType.MYSQL.equals(dbType) || DbType.MARIADB.equals(dbType) ) {
            return false;
        }

        boolean globalConditionalInsertEnabled = dbLockConditionalInsertEnabled.getBoolean();
        log.debug("Lock conditional insert property is enabled: {}", globalConditionalInsertEnabled);
        if (globalConditionalInsertEnabled && DbType.POSTGRESQL.equals(dbType)) {
            return isPostgresVersionSupportConditionalInsert();
        }
        return globalConditionalInsertEnabled;
    }

    /**
     * Check if PostgreSQL db supports conditional insert (supported since 9.5),
     * should be called with for PostgreSQL only
     */
    @VisibleForTesting
    boolean isPostgresVersionSupportConditionalInsert() {
        log.debug("Checking if conditional lock is enabled for PostgreSQL DB");
        if (DbType.POSTGRESQL.equals(dbType)) {
            long postgresMinVersion = 90500;
            try (Connection conn = getLocksConnection()) {
                long dbVersion = getPostgreSQLVersionNumber(conn);
                log.debug("PostgreSQL version is: {}", dbVersion);
                return dbVersion >= postgresMinVersion;
            } catch (Exception e) {
                log.error("Unable to detect database version {}", e.getMessage());
                log.debug("Unable to detect database version, assuming that database supports `ON CONFLICT` clause " +
                        "for conditional lock inserts", e);
                return false;
            }
        }
        // assuming that the undetected database supports conditional insert
        return true;
    }

    private int getPostgreSQLVersionNumber(Connection con) throws SQLException {
        int version = 0;
        try (PreparedStatement preparedStatement = con.prepareStatement("show server_version_num;")) {
            ResultSet rs = preparedStatement.executeQuery();
            if (rs.next()) {
                version = rs.getInt(1);
            }
        }
        return version;
    }
}