package org.artifactory.storage.db.event.service;

import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.events.EventLogHintManager;
import org.artifactory.event.EventHintType;
import org.artifactory.event.EventType;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.sapi.fs.VfsItem;
import org.artifactory.storage.event.EventInfo;
import org.artifactory.storage.event.EventsService;
import org.artifactory.storage.event.ShiftEventsService;
import org.mockito.ArgumentCaptor;
import org.mockito.Mockito;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.List;

/**
 * @author shaharz
 */
@Test
public class EventsSessionResourceTest {

    private ShiftEventsService shiftEventsService;
    private EventsService eventsService;

    @BeforeMethod
    private void setup(){
        ArtifactoryContext context = Mockito.mock(ArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(context);
        shiftEventsService = Mockito.mock(ShiftEventsService.class);
        eventsService = Mockito.mock(EventsService.class);
        Mockito.when(context.beanForType(Mockito.eq(ShiftEventsService.class)))
                .thenReturn(shiftEventsService);
        Mockito.when(context.beanForType(Mockito.eq(EventsService.class)))
                .thenReturn(eventsService);
        Mockito.when(eventsService.isEnabled()).thenReturn(true);
    }

    @AfterMethod
    private void destroy() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void testBeforeCommit(){
        EventLogHintManager.put(EventHintType.FederatedReplication, "bla");
        EventsSessionResource session = new EventsSessionResource();
        VfsItem item = Mockito.mock(VfsItem.class);
        Mockito.when(item.getRepoPath()).thenReturn(new RepoPathImpl("repo1", "/a/b"));
        session.itemCreated(item);
        session.beforeCommit();

        ArgumentCaptor<List<EventInfo>> events = ArgumentCaptor.forClass(List.class);
        Mockito.verify(eventsService, Mockito.times(1)).appendEvents(events.capture());

        Assert.assertEquals(events.getValue().size(), 1);
        Assert.assertEquals(events.getValue().get(0).getPath(), "repo1/a/b");
        Assert.assertEquals(events.getValue().get(0).getType(), EventType.create);
        Assert.assertEquals(events.getValue().get(0).getHints().size(), 1);
    }

    @Test (dataProvider = "AfterCompletionProvider")
    public void testAfterCompletionCommitTrueWithEvents(boolean commit, boolean hasEvents, int expectedShifts){
        EventLogHintManager.put(EventHintType.FederatedReplication, "bla");
        EventsSessionResource session = new EventsSessionResource();

        if (hasEvents) {
            VfsItem item = Mockito.mock(VfsItem.class);
            Mockito.when(item.getRepoPath()).thenReturn(new RepoPathImpl("repo1", "/a/b"));
            session.itemCreated(item);
            session.beforeCommit();
        }
        session.afterCompletion(commit);
        Mockito.verify(shiftEventsService, Mockito.times(expectedShifts)).triggerShift();
    }

    @DataProvider(name = "AfterCompletionProvider")
    private Object[][] testAfterCompletionProvider(){
        return new Object[][]{
                {false, false, 0},
                {false, true, 0},
                {true, false, 0},
                {true, true, 1}
        };
    }
}
